import logging
from collections import defaultdict
from ipaddress import ip_address

from pihole6api import PiHole6Client

from octodns.provider.base import BaseProvider
from octodns.record import Record


class PiholeProvider(BaseProvider):
    DEFAULT_TTL = 86400  # TTL does not matter/unsupported for Pi-hole

    SUPPORTS_GEO = False
    SUPPORTS_DYNAMIC = False
    SUPPORTS_ROOT_NS = False
    SUPPORTS = set(('A', 'AAAA', 'CNAME'))

    def __init__(self, id, url, password, *args, **kwargs):
        self.log = logging.getLogger(f'PiholeProvider[{id}]')
        self.log.debug('__init__: id=%s, url=%s', id, url)
        super().__init__(id, *args, **kwargs)

        self._client = PiHole6Client(url, password)

    def _data_for_multiple(self, type, records):
        return {
            'ttl': PiholeProvider.DEFAULT_TTL,
            'type': type,
            'values': [r for r in records],
        }

    _data_for_A = _data_for_multiple
    _data_for_AAAA = _data_for_multiple

    def _data_for_CNAME(self, _type, records):
        values = records[0].split(',')

        return {
            'ttl': int(values[1]) or PiholeProvider.DEFAULT_TTL,
            'type': _type,
            'value': f'{values[0]}',
        }

    def _process_desired_zone(self, desired):
        # TTL is not supported for records in Pi-hole.
        # Reset desired records TTL to a known default to
        # ignore the source provider and prevent false changes.
        for record in desired.records:
            record = record.copy()
            if record._type != 'CNAME':
                record.ttl = PiholeProvider.DEFAULT_TTL
            desired.add_record(record, replace=True)

        return super()._process_desired_zone(desired)

    def populate(self, zone, target=False, lenient=False):
        self.log.debug(
            'populate: name=%s, target=%s, lenient=%s',
            zone.name,
            target,
            lenient,
        )

        values = defaultdict(lambda: defaultdict(list))
        pihole_config = self._client.config.get_config()
        # A/AAAA "records"
        for entry in pihole_config['config']['dns']['hosts']:
            ip, name = entry.split(' ', 1)
            # Pi-hole does not really have the concept of zones
            # we only want to return "records" within the zone
            if zone.name not in name:
                continue

            # Strip the zone name from the list entry
            name = name.split(zone.name, 1)[0].rstrip('.')

            match ip_address(ip).version:
                case 4:
                    values[name]['A'].append(ip)
                case 6:  # pragma: no cover (this is tested but coverage warns)
                    values[name]['AAAA'].append(ip)

        # CNAME "records"
        for entry in pihole_config['config']['dns']['cnameRecords']:
            name, target = entry.split(',', 1)

            # Pi-hole does not really have the concept of zones
            # we only want to return "records" within the zone
            if zone.name not in name:
                continue

            # Strip the zone name from the list entry
            name = name.split(f".{zone.name}", 1)[0]
            values[name]['CNAME'].append(target)

        before = len([r for r in values.values()])
        for name, types in values.items():
            for _type, records in types.items():
                data_for = getattr(self, f'_data_for_{_type}')
                record = Record.new(
                    zone,
                    name,
                    data_for(_type, records),
                    source=self,
                    lenient=lenient,
                )
                zone.add_record(record, lenient=lenient)

        exists = len(values) > 0
        self.log.info(
            'populate:   found %s records, exists=%s',
            len(zone.records) - before,
            exists,
        )
        return exists

    def _params_for_multiple(self, record):
        for value in record.values:
            yield {
                'name': (
                    f"{record.name}.{record.zone.name}"
                    if record.name
                    else record.zone.name
                ),
                'data': value,
            }

    _params_for_A = _params_for_multiple
    _params_for_AAAA = _params_for_multiple

    def _params_for_single(self, record):
        yield {
            'name': (
                f"{record.name}.{record.zone.name}"
                if record.name
                else record.zone.name
            ),
            'ttl': record.ttl,
            'data': record.value,
        }

    _params_for_CNAME = _params_for_single

    def _apply_Create(self, change):
        new = change.new
        params_for = getattr(self, f'_params_for_{new._type}')
        for params in params_for(new):
            result = None
            match change.record._type:
                case 'A' | 'AAAA':
                    result = self._client.config.add_local_a_record(
                        params['name'], params['data']
                    )
                case 'CNAME':
                    result = self._client.config.add_local_cname(
                        params['name'], params['data'], ttl=params['ttl']
                    )

            if 'error' in result or result is None:
                raise ValueError(
                    f"Failed to apply change for record: {change.record.name}"
                )

    def _apply_Update(self, change):
        self._apply_Delete(change)
        self._apply_Create(change)

    def _apply_Delete(self, change):
        existing = change.existing
        params_for = getattr(self, f'_params_for_{existing._type}')
        for params in params_for(existing):
            result = None
            match change.record._type:
                case 'A' | 'AAAA':
                    result = self._client.config.remove_local_a_record(
                        params['name'], params['data']
                    )
                case 'CNAME':
                    result = self._client.config.remove_local_cname(
                        params['name'], params['data'], ttl=params['ttl']
                    )

            if 'error' in result or result is None:
                raise ValueError(
                    f"Failed to delete the record: {change.record.name}"
                )

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        self.log.debug(
            '_apply: zone=%s, len(changes)=%d', desired.name, len(changes)
        )

        for change in changes:
            class_name = change.__class__.__name__
            getattr(self, f'_apply_{class_name}')(change)

        self.log.info('_apply: sending changes to Pi-hole')



from unittest.mock import MagicMock, patch

import pytest

from octodns.provider.plan import Plan
from octodns.record import Record
from octodns.record.change import Create, Delete, Update
from octodns.zone import Zone

from octodns_pihole import PiholeProvider

ZONE_NAME = "unit.tests."


def _get_provider(mock_client):
    with patch('octodns_pihole.PiHole6Client', return_value=mock_client):
        return PiholeProvider('test', 'http://pihole.mock', 'password')


def _get_zone():
    return Zone(ZONE_NAME, [])


def _make_config(hosts=None, cnames=None):
    return {
        "config": {
            "dns": {
                "hosts": hosts if hosts is not None else [],
                "cnameRecords": (cnames if cnames is not None else []),
            }
        }
    }


def _populate(mock_client, config):
    mock_client.config.get_config.return_value = config
    provider = _get_provider(mock_client)
    zone = _get_zone()
    provider.populate(zone)
    return zone, provider


def _mock_client_for_apply():
    mock_client = MagicMock()
    mock_client.config.get_config.return_value = _make_config()
    mock_client.config.add_local_a_record.return_value = {}
    mock_client.config.add_local_cname.return_value = {}
    mock_client.config.remove_local_a_record.return_value = {}
    mock_client.config.remove_local_cname.return_value = {}
    return mock_client


# --- Tests: populate ---


def test_populate_empty():
    zone, _ = _populate(MagicMock(), _make_config())
    assert len(zone.records) == 0


def test_populate_A():
    config = _make_config(
        hosts=["1.2.3.4 www.unit.tests.", "1.2.3.5 www.unit.tests."]
    )
    zone, _ = _populate(MagicMock(), config)
    record = list(zone.records)[0]
    assert record._type == "A"
    assert record.name == "www"
    assert sorted(record.values) == ["1.2.3.4", "1.2.3.5"]
    assert record.ttl == 86400


def test_populate_A_root():
    config = _make_config(hosts=["1.2.3.4 unit.tests."])
    zone, _ = _populate(MagicMock(), config)
    record = list(zone.records)[0]
    assert record._type == "A"
    assert record.name == ""
    assert list(record.values) == ["1.2.3.4"]


def test_populate_AAAA():
    config = _make_config(
        hosts=["2601:644:500:e210:62f8:1dff:feb8:947a aaaa.unit.tests."]
    )
    zone, _ = _populate(MagicMock(), config)
    record = list(zone.records)[0]
    assert record._type == "AAAA"
    assert record.name == "aaaa"
    assert list(record.values) == ["2601:644:500:e210:62f8:1dff:feb8:947a"]


def test_populate_CNAME():
    config = _make_config(cnames=["cname.unit.tests.,target.unit.tests.,300"])
    zone, _ = _populate(MagicMock(), config)
    record = list(zone.records)[0]
    assert record._type == "CNAME"
    assert record.name == "cname"
    assert record.value == "target.unit.tests."
    assert record.ttl == 300


def test_populate_filters_by_zone():
    config = _make_config(
        hosts=["1.1.1.1 other.tld.", "1.2.3.4 www.unit.tests."],
        cnames=["alias.other.tld.,target.other.tld.,300"],
    )
    zone, _ = _populate(MagicMock(), config)
    assert len(zone.records) == 1
    record = list(zone.records)[0]
    assert record.name == "www"


def test_populate_all():
    config = _make_config(
        hosts=[
            "1.2.3.4 unit.tests.",
            "1.2.3.5 unit.tests.",
            "2.2.3.6 www.unit.tests.",
            "2601:644:500:e210:62f8:1dff:feb8:947a aaaa.unit.tests.",
        ],
        cnames=["cname.unit.tests.,unit.tests.,300"],
    )
    zone, _ = _populate(MagicMock(), config)
    type_names = {r._type for r in zone.records}
    assert type_names == {"A", "AAAA", "CNAME"}
    # root A (2 values) + www A + aaaa AAAA + cname CNAME
    assert len(zone.records) == 4


# --- Tests: _apply create ---


def test_apply_create_A():
    mock_client = _mock_client_for_apply()
    provider = _get_provider(mock_client)

    zone = _get_zone()
    record = Record.new(
        zone,
        "www",
        {"type": "A", "ttl": 86400, "values": ["1.2.3.4", "5.6.7.8"]},
    )

    change = Create(record)
    plan = Plan(zone, zone, [change], True)
    provider._apply(plan)

    assert mock_client.config.add_local_a_record.call_count == 2
    mock_client.config.add_local_a_record.assert_any_call(
        "www.unit.tests.", "1.2.3.4"
    )
    mock_client.config.add_local_a_record.assert_any_call(
        "www.unit.tests.", "5.6.7.8"
    )


def test_apply_create_root_A():
    mock_client = _mock_client_for_apply()
    provider = _get_provider(mock_client)

    zone = _get_zone()
    record = Record.new(
        zone, "", {"type": "A", "ttl": 86400, "value": "1.2.3.4"}
    )

    change = Create(record)
    plan = Plan(zone, zone, [change], True)
    provider._apply(plan)

    mock_client.config.add_local_a_record.assert_called_once_with(
        "unit.tests.", "1.2.3.4"
    )


def test_apply_create_AAAA():
    mock_client = _mock_client_for_apply()
    provider = _get_provider(mock_client)

    zone = _get_zone()
    record = Record.new(
        zone, "ipv6", {"type": "AAAA", "ttl": 86400, "values": ["2001:db8::1"]}
    )

    change = Create(record)
    plan = Plan(zone, zone, [change], True)
    provider._apply(plan)

    mock_client.config.add_local_a_record.assert_called_once_with(
        "ipv6.unit.tests.", "2001:db8::1"
    )


def test_apply_create_CNAME():
    mock_client = _mock_client_for_apply()
    provider = _get_provider(mock_client)

    zone = _get_zone()
    record = Record.new(
        zone, "alias", {"type": "CNAME", "ttl": 300, "value": "www.unit.tests."}
    )

    change = Create(record)
    plan = Plan(zone, zone, [change], True)
    provider._apply(plan)

    mock_client.config.add_local_cname.assert_called_once_with(
        "alias.unit.tests.", "www.unit.tests.", ttl=300
    )


# --- Tests: _apply delete ---


def test_apply_delete_A():
    mock_client = _mock_client_for_apply()
    provider = _get_provider(mock_client)

    zone = _get_zone()
    existing = Record.new(
        zone,
        "www",
        {"type": "A", "ttl": 86400, "values": ["1.2.3.4", "5.6.7.8"]},
    )

    change = Delete(existing)
    plan = Plan(zone, zone, [change], True)
    provider._apply(plan)

    assert mock_client.config.remove_local_a_record.call_count == 2
    mock_client.config.remove_local_a_record.assert_any_call(
        "www.unit.tests.", "1.2.3.4"
    )
    mock_client.config.remove_local_a_record.assert_any_call(
        "www.unit.tests.", "5.6.7.8"
    )


def test_apply_delete_CNAME():
    mock_client = _mock_client_for_apply()
    provider = _get_provider(mock_client)

    zone = _get_zone()
    existing = Record.new(
        zone, "alias", {"type": "CNAME", "ttl": 300, "value": "www.unit.tests."}
    )

    change = Delete(existing)
    plan = Plan(zone, zone, [change], True)
    provider._apply(plan)

    mock_client.config.remove_local_cname.assert_called_once_with(
        "alias.unit.tests.", "www.unit.tests.", ttl=300
    )


# --- Tests: _apply update ---


def test_apply_update_A():
    mock_client = _mock_client_for_apply()
    provider = _get_provider(mock_client)

    zone = _get_zone()
    existing = Record.new(
        zone, "www", {"type": "A", "ttl": 86400, "value": "1.2.3.4"}
    )
    new = Record.new(
        zone, "www", {"type": "A", "ttl": 86400, "value": "9.8.7.6"}
    )

    change = Update(existing, new)
    plan = Plan(zone, zone, [change], True)
    provider._apply(plan)

    # Delete old
    mock_client.config.remove_local_a_record.assert_called_once_with(
        "www.unit.tests.", "1.2.3.4"
    )
    # Create new
    mock_client.config.add_local_a_record.assert_called_once_with(
        "www.unit.tests.", "9.8.7.6"
    )


# --- Tests: error handling ---


def test_apply_create_raises_on_error():
    mock_client = _mock_client_for_apply()
    mock_client.config.add_local_a_record.return_value = {
        "error": "something went wrong"
    }
    provider = _get_provider(mock_client)

    zone = _get_zone()
    record = Record.new(
        zone, "www", {"type": "A", "ttl": 86400, "value": "1.2.3.4"}
    )

    change = Create(record)
    plan = Plan(zone, zone, [change], True)

    with pytest.raises(ValueError):
        provider._apply(plan)


def test_apply_delete_raises_on_error():
    mock_client = _mock_client_for_apply()
    mock_client.config.remove_local_a_record.return_value = {
        "error": "not found"
    }
    provider = _get_provider(mock_client)

    zone = _get_zone()
    existing = Record.new(
        zone, "www", {"type": "A", "ttl": 86400, "value": "1.2.3.4"}
    )

    change = Delete(existing)
    plan = Plan(zone, zone, [change], True)

    with pytest.raises(ValueError):
        provider._apply(plan)
